from .base_container import BaseContainer


class Misc(BaseContainer):
    """
    This is used for modules that don't
    have any optimizable parameters, and other
    miscellaneous objects.
    """

    pass
